package de.unifreiburg.twodeedoo.view;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import javax.swing.JComponent;
import javax.swing.Timer;



/**
 * I'm a Swing widget, I feed an {@link IActivity} with events and 
 * have a {@link IPainter} paint my window at regular intervals. 
 * <p>
 * Caution: I own a {@link Timer} which creates a thread when running,
 * so be sure to call {@link #stop()} during shutdown.
 * </p>
 * <p>
 * You probably want to register a {@link IGameControlListener}
 * to be notified when the activity wants the game to end. 
 * </p>
 * @author anton
 *
 */
public class SceneView extends JComponent implements IGameController {
	
	/**
	 * Initial delay of timer in ms. Will be changed after first refresh.
	 */
	private static final int INITIAL_DELAY = 15;
	/**
	 * For serialization. Eclipse complains if I lack this.
	 */
	private static final long serialVersionUID = 1L;
	/**
	 * The activity. Will be called to update itself for each update,
	 * and will receive key events.
	 */
	private IActivity activity;

	/**
	 * The scene, or something which paints itself onto a {@link Graphics}.
	 */
	private IPainter scene;
	
	/**
	 * The timer to schedule the next model update.
	 */
	private Timer timer;

	/**
	 * Wall-clock time of the last model update.
	 */
	private long millisLastRun=0;
	
	/**
	 * Width of the display area
	 */
	private int width;
	
	/**
	 * Height of the display area
	 */
	private int height;
	
	private Queue<InputEvent> inputQueue = new LinkedList<InputEvent>();
	
	/**
	 * True from when you {@link #switchToActivity(IActivity)} to
	 * when I schedule the next activity execution.
	 */
	private boolean activitySwitched;
	
	/**
	 * Listeners for major game-changing events
	 */
	private List<IGameControlListener> gameControlListeners = 
		new LinkedList<IGameControlListener>();
	
	/**
	 * Init. This installs dummy instances for {@link #activity}
	 * and {@link #scene}. Replace them using setters when you feel like it.
	 */
	public SceneView(int width, int height){
		this.width = width;
		this.height = height;
		this.activity = new NullActivity();
		this.scene = IPainter.NULL_OBJECT;
	}

	/**
	 * Swing demands a repainting. Paint the scene and schedule the next update.
	 */
	@Override
	public void paint(Graphics g) {
		scene.paint(g);
		restartTimerIfExpired();
	}

	/**
	 * Start the timer for regular model updates.
	 */
	public void initTimer() {
		millisLastRun = System.currentTimeMillis();
		this.timer = new Timer(INITIAL_DELAY, new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				long millisNow = System.currentTimeMillis();
				int elapsedTime = (int) (millisNow - millisLastRun);
				millisLastRun = millisNow;
				activitySwitched = false;
				feedInputEvents();
				activity.run(elapsedTime, SceneView.this);
				repaint();
			}
		});
		timer.setRepeats(false);
		timer.start();
	}

	protected void feedInputEvents() {
		for(InputEvent ie : inputQueue){
			ie.sendTo(activity, this);
			
			if(activitySwitched)
				break; // stop delivering events to the activity
		}
		inputQueue.clear();
	}

	/**
	 * Schedule the next update so that we get as many FPS as possible,
	 * but without refreshing earlier than {@link #MIN_DELAY} after the last time.
	 */
	private void restartTimerIfExpired() {
		if(timer.isRunning()) {
			// if the timer is still running, no need to restart it.
			return;
		}
		long millisNow = System.currentTimeMillis();
		
		// Remaining wait time until earliest allowed time for next update
		// This addition may overflow once, long after 3000. I don't care.
		long remainingTime = (millisLastRun + getMinDelay()) - millisNow; 
		if(remainingTime < 0) {
			// wow we're already late. Let's schedule for now+1ms.
			remainingTime = 1;
		}
		timer.setDelay((int) remainingTime);
		timer.restart(); 
	}

	private int getMinDelay() {
		return this.activity.minimalTimeSlice();
	}

	/**
	 * Stop the timer, so that we may get rid of the timer thread.
	 * Call this before closing the window.
	 */
	public void stop(){
		timer.stop();
	}
	

	/**
	 * Prepare for keyboard input. Window must be visible.
	 */
	public void initKey() {
		requestFocus();
		// use an ad-hoc adapter to filter AWT key events down to Twodeedoo key events
		addKeyListener(new KeyListener() {
			/** Ignore */
			@Override
			public void keyTyped(KeyEvent e) {
			}
			
			/** Dispatch to activity */
			@Override
			public void keyReleased(KeyEvent e) {
				inputQueue.add(InputEvent.makeReleased(e.getKeyCode()));
			}
			
			/** Dispatch to activity */
			@Override
			public void keyPressed(KeyEvent e) {
				inputQueue.add(InputEvent.makePressed(e.getKeyCode()));
			}
		});
	}

	public void addGameControlListener(IGameControlListener listener) {
		this.gameControlListeners.add(listener);
	}
	
	public void removeGameControlListener(IGameControlListener listener) {
		this.gameControlListeners.remove(listener);
	}
	
	/**
	 * Set the scene
	 * @param scene something that can paint to a {@link Graphics}, usually an IScene.
	 */
	public void setScene(IPainter scene) {
		this.scene = scene;
	}

	/**
	 * How big do we want the window to be?
	 */
	@Override
	public Dimension getPreferredSize() {
		return new Dimension(width, height);
	}


	@Override
	public void switchToActivity(IActivity newActivity) {
		this.activity = newActivity;
		activitySwitched = true;
	}

	
	
	@Override
	public void quitGame() {
		// This is called through IGameController; just tell the listeners,
		// hoping the game application is among them.
		this.stop();
		ArrayList<IGameControlListener> listenersCopy = 
			new ArrayList<IGameControlListener>(gameControlListeners);
		for(IGameControlListener listener :	listenersCopy) {
			listener.quitRequested();
		}
	}

	@Override
	public void restartGame() {
		// This is called through IGameController; just tell the listeners,
		// hoping the game application is among them.
		ArrayList<IGameControlListener> listenersCopy = 
			new ArrayList<IGameControlListener>(gameControlListeners);
		for(IGameControlListener listener :	listenersCopy) {
			listener.restartRequested();
		}
	}
	
	@Override
	public boolean isActivitySwitched() {
		return activitySwitched;
	}

	/**
	 * @return the width
	 */
	public int getWidth() {
		return width;
	}

	/**
	 * @return the height
	 */
	public int getHeight() {
		return height;
	}
}
